#define ARRAY_MT "array-mt"

#define LUA_LIB
#include <lua.h>
#include <lauxlib.h>

static int array_tostring(lua_State *L) {
    luaL_checkudata(L, 1, ARRAY_MT);
    lua_pushfstring(L, "array[%d]: %p", lua_getuvcount(L, 1), lua_touserdata(L, 1));
    return 1;
}

static int array_len(lua_State *L) {
    luaL_checkudata(L, 1, ARRAY_MT);
    lua_pushinteger(L, lua_getuvcount(L, 1));
    return 1;
}

static int ipairsaux(lua_State *L) {
  int i = luaL_checkint(L, 2);
  luaL_checkudata(L, 1, ARRAY_MT);
  i++;  /* next value */
  if (i > lua_getuvcount(L, 1))
      return 0;
  lua_pushinteger(L, i);
  lua_getuservalueev(L, 1, i);
  return 2;
}

static int array_ipairs(lua_State *L) {
    luaL_checkudata(L, 1, ARRAY_MT);
    lua_pushcfunction(L, ipairsaux);
    lua_pushvalue(L, 1);
    lua_pushinteger(L, 0);
    return 3;
}

static int array_index(lua_State *L) {
    int idx, count;
    luaL_checkudata(L, 1, ARRAY_MT);
    idx = luaL_checkint(L, 2);
    count = lua_getuvcount(L, 1);
    if (idx < 0) idx += count + 1;
    if (idx >= 1 && idx <= count)
    {
        lua_getuservalueev(L, 1, idx);
        return 1;
    }
    return 0;
}

static int array_newindex(lua_State *L) {
    int idx, count;
    luaL_checkudata(L, 1, ARRAY_MT);
    idx = luaL_checkint(L, 2);
    count = lua_getuvcount(L, 1);
    if (idx < 0) idx += count + 1;
    if (idx >= 1 && idx <= count)
        lua_setuservalueev(L, 1, idx);
    return 0;
}

static int array_new(lua_State *L) {
    if (lua_gettop(L) == 1 && lua_isnumber(L, 1))
        lua_newuserdataev(L, 0, lua_tonumber(L, 1));
    else
    {
        int i, len = lua_gettop(L);
        lua_newuserdataev(L, 0, len);

        for (i = 1; i <= len; ++i)
        {
            lua_pushvalue(L, i);
            lua_setuservalueev(L, -2, i);
        }
    }
    luaL_setmetatable(L, ARRAY_MT);
    return 1;
}

static luaL_Reg array_mts[] = {
    {"__tostring", array_tostring},
    {"__len", array_len},
    {"__ipairs", array_ipairs},
    {"__index", array_index},
    {"__newindex", array_newindex},
};

static luaL_Reg array_funcs[] = {
    {"new", array_new},
};

LUALIB_API int luaopen_array(lua_State *L) {
    luaL_newmetatable(L, ARRAY_MT);
    luaL_setfuncs(L, array_mts, 0);
    luaL_newlib(L, array_funcs);
    return 1;
}
